# This file is part of SSHare.
#
# SSHare is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# SSHare is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# SSHare. If not, see <https://www.gnu.org/licenses/>.

import argparse
import getpass
import importlib
import importlib.util
import os
import os.path
import pyclip
import time
import tomllib
import subprocess
import sys
from pathlib import Path
from version import version

from plugins.config import Default
from plugins.config import NoDefault
from plugins.config import Flags

class Congloggerate:
    def __init__(self, loggers):
        def info(message):
            for logger in loggers:
                logger.info(message)

        def warn(message):
            for logger in loggers:
                logger.warn(message)

        def error(message):
            for logger in loggers:
                logger.error(message)

        self.info  = info
        self.warn  = warn
        self.error = error

        fatalicize(self)

def fatalicize(logger):
    def fatal(message):
        logger.error(message)
        sys.exit(1)
    setattr(logger, "fatal", fatal)

class Plugin:
    def __init__(self, name, module):
        self.name   = name
        self.module = module

def main():
    config_directory = Path(os.environ.get("XDG_CONFIG_DIR") or f"{os.environ["HOME"]}/.config") / "sshare"
    with open(config_directory / "config.toml", mode="rb") as file:
        config = tomllib.load(file)

    # Load command line early and set it as the active logger
    # so that it can be used to report errors while loading and
    # configuring plugins
    # i.e. before other logging plugins have had a chance to be initialised
    logger = importlib.import_module("plugins.default.command_line")
    fatalicize(logger)

    # Load inbuilt plugins
    plugins_flat = [
        Plugin("command_line", logger),
        Plugin("file",         importlib.import_module("plugins.default.file")),
        Plugin("current_time", importlib.import_module("plugins.default.current_time")),
        Plugin("append_type",  importlib.import_module("plugins.default.append_type")),
        Plugin("ssh",          importlib.import_module("plugins.default.ssh")),
    ]
    plugins = {
        "logger": [],
        "source": [],
        "name":   [],
        "upload": [],
    }

    # Load external plugins
    sys.dont_write_bytecode = True
    for plugin in (config_directory / "plugins").iterdir():
        if plugin.is_file():
            module_spec = importlib.util.spec_from_file_location(
                plugin.stem,
                plugin.as_posix(),
            )
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)
            plugins_flat.append(Plugin(plugin.stem, module))
    sys.dont_write_bytecode = False

    # Set plugin configurations from config file
    if config.get("plugins") == None:
        config["plugins"] = {}
    for plugin in plugins_flat:
        if hasattr(plugin.module, "config"):
            plugin_config = config["plugins"].get(plugin.name)
            if plugin_config != None:
                for config_entry in plugin_config.items():
                    plugin.module.config[config_entry[0]] = Default(config_entry[1])

    # Flatten plugin configs
    class PluginConfig: pass
    error = False
    for plugin in plugins_flat:
        if hasattr(plugin.module, "config"):
            config = plugin.module.config
            plugin.module.config = PluginConfig()
            for config_entry in config.items():
                if isinstance(config_entry[1], NoDefault):
                    logger.error(f"{plugin.name} > Error: Value '{config_entry[0]}' has no default value and must be specified explicitly")
                    error = True
                elif isinstance(config_entry[1], Default):
                    setattr(plugin.module.config, config_entry[0], config_entry[1].value)
                else:
                    setattr(plugin.module.config, config_entry[0], config_entry[1])
    if error:
        sys.exit(1)

    # Initialise plugins
    for plugin in plugins_flat:
        setattr(plugin.module, "logger", logger)
        if hasattr(plugin.module, "init"):
            plugin.module.init()

    # Sort plugins by type
    error = False
    for plugin in plugins_flat:
        if isinstance(plugin.module.plugin_type, str):
            plugin.module.plugin_type = [ plugin.module.plugin_type ]
        for plugin_type in plugin.module.plugin_type:
            plugins_of_type = plugins.get(plugin_type)
            if plugins_of_type == None:
                logger.error(f"Error: Plugin '{plugin.name}' has an invalid plugin type '{plugin_type}'")
                error = True
            else:
                plugins_of_type.append(plugin)
    if error:
        sys.exit(1)

    logger = Congloggerate([ logger.module for logger in plugins["logger"] ])

    sources = []
    for plugin in plugins["source"]:
        sources.append(plugin.module.get())

    for index, source in enumerate(sources):
        name = ""
        for plugin in plugins["name"]:
            name = plugin.module.name(name, source)
        sources[index] = name, source

    for (name, source) in sources:
        for plugin in plugins["upload"]:
            plugin.module.upload(name, source)

    sys.exit(0)

def parse_arguments():
    parser = argparse.ArgumentParser(
        prog = "SSHare",
        description = "Upload files to a server via ssh",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s version {version}",
    )
    parser.add_argument(
        "-l",
        "--latest",
        action="store_const",
        const=True,
        help="Upload the latest image from the source directory",
    )
    parser.add_argument(
        "-p",
        "--paste",
        action="store_const",
        const=True,
        help="Upload the contents of the clipboard as a .txt file",
    )
    parser.add_argument(
        "-f",
        "--file",
        help="Upload a file",
    )
    parser.add_argument(
        "-c",
        "--copy",
        action="store_const",
        const=True,
        help="Copy the resultant URL to the clipboard",
    )
    arguments = parser.parse_args()

    return arguments


def _latest(directory, key=os.path.getmtime):
    files         = map(lambda file: f"{directory}/{file}", os.listdir(directory))
    selection     = next(files)
    selection_key = key(selection)
    for file in files:
        new_key = key(file)
        if new_key > selection_key:
            selection     = file
            selection_key = key
    return selection
